@@ -1,17 +1,5 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-"""
-Author: Farid Mohammadi, M.Sc.
-E-Mail: farid.mohammadi@iws.uni-stuttgart.de
-Department of Hydromechanics and Modelling of Hydrosystems (LH2)
-Institute for Modelling Hydraulic and Environmental Systems (IWS), University
-of Stuttgart, www.iws.uni-stuttgart.de/lh2/
-Pfaffenwaldring 61
-70569 Stuttgart
-
-Created on Sat Aug 24 2019
-
-"""
 
 import numpy as np
 import math
@@ -24,15 +12,120 @@ from .apoly_construction import apoly_construction
 
 
 class ExpDesigns:
-    def __init__(self, Input, meta_Model='pce', hdf5_file=None,
+    """
+    This class generates samples from the prescribed marginals for the model
+    parameters using the `Input` object.
+
+    Attributes
+    ----------
+    Input : obj
+        Input object containing the parameter marginals, i.e. name,
+        distribution type and distribution parameters or available raw data.
+    method : str
+        Type of the experimental design. The default is `'normal'`. Other
+        option is `'sequential'`.
+    meta_Model : str
+        Type of the meta_model.
+    sampling_method : str
+        Name of the sampling method for the experimental design. The following
+        sampling method are supported:
+
+        * random
+        * latin_hypercube
+        * sobol
+        * halton
+        * hammersley
+        * korobov
+        * chebyshev(FT)
+        * grid(FT)
+        * nested_grid(FT)
+        * user
+    hdf5_file : str
+        Name of the hdf5 file that contains the experimental design.
+    n_new_samples : int
+        Number of (initial) training points.
+    n_max_samples : int
+        Number of maximum training points.
+    mod_LOO_threshold : float
+        The modified leave-one-out cross validation threshold where the
+        sequential design stops.
+    tradeoff_scheme : str
+        Trade-off scheme to assign weights to the exploration and exploitation
+        scores in the sequential design.
+    n_canddidate : int
+        Number of candidate training sets to calculate the scores for.
+    explore_method : str
+        Type of the exploration method for the sequential design. The following
+        methods are supported:
+
+        * Voronoi
+        * random
+        * latin_hypercube
+        * LOOCV
+        * dual annealing
+    exploit_method : str
+        Type of the exploitation method for the sequential design. The
+        following methods are supported:
+
+        * BayesOptDesign
+        * BayesActDesign
+        * VarOptDesign
+        * alphabetic
+        * Space-filling
+    util_func : str or list
+        The utility function to be specified for the `exploit_method`. For the
+        available utility functions see Note section.
+    n_cand_groups : int
+        Number of candidate groups. Each group of candidate training sets will
+        be evaulated separately in parallel.
+    n_replication : int
+        Number of replications. Only for comparison. The default is 1.
+    post_snapshot : int
+        Whether to plot the posterior in the sequential design. The default is
+        `True`.
+    step_snapshot : int
+        The number of steps to plot the posterior in the sequential design. The
+        default is 1.
+    max_a_post : list or array
+        Maximum a posteriori of the posterior distribution, if known. The
+        default is `[]`.
+    adapt_verbose : bool
+        Whether to plot the model response vs that of metamodel for the new
+        trining point in the sequential design.
+
+    Note
+    ----------
+    The following utiliy functions for the **exploitation** methods are
+    supported:
+
+    #### BayesOptDesign (when data is available)
+    - DKL (Kullback-Leibler Divergence)
+    - DPP (D-Posterior-percision)
+    - APP (A-Posterior-percision)
+
+    #### VarBasedOptDesign -> when data is not available
+    - Entropy (Entropy/MMSE/active learning)
+    - EIGF (Expected Improvement for Global fit)
+    - LOOCV (Leave-one-out Cross Validation)
+
+    #### alphabetic
+    - D-Opt (D-Optimality)
+    - A-Opt (A-Optimality)
+    - K-Opt (K-Optimality)
+    """
+
+    def __init__(self, Input, method='normal', meta_Model='pce',
+                 sampling_method='random', hdf5_file=None,
                  n_new_samples=1, n_max_samples=None, mod_LOO_threshold=1e-16,
                  tradeoff_scheme=None, n_canddidate=1, explore_method='random',
                  exploit_method='Space-filling', util_func='Space-filling',
                  n_cand_groups=4, n_replication=1, post_snapshot=False,
-                 step_snapshot=1, max_a_post=[]):
+                 step_snapshot=1, max_a_post=[], adapt_verbose=False):
 
         self.InputObj = Input
+        self.method = method
         self.meta_Model = meta_Model
+        self.sampling_method = sampling_method
         self.hdf5_file = hdf5_file
         self.n_new_samples = n_new_samples
         self.n_max_samples = n_max_samples
@@ -47,6 +140,7 @@ class ExpDesigns:
         self.post_snapshot = post_snapshot
         self.step_snapshot = step_snapshot
         self.max_a_post = max_a_post
+        self.adapt_verbose = adapt_verbose
 
     # -------------------------------------------------------------------------
     def generate_samples(self, n_samples, sampling_method='random',
@@ -58,11 +152,11 @@ class ExpDesigns:
         ----------
         n_samples : int
             Number of requested samples.
-        sampling_method : TYPE, optional
-            DESCRIPTION. The default is 'random'.
+        sampling_method : str, optional
+            Sampling method. The default is `'random'`.
         transform : bool, optional
             Transformation via an isoprobabilistic transformation method. The
-            default is False.
+            default is `False`.
 
         Returns
         -------
@@ -71,10 +165,11 @@ class ExpDesigns:
 
         """
         try:
-            samples = chaospy.generate_samples(n_samples, domain=self.JDist,
-                                               rule=sampling_method)
+            samples = chaospy.generate_samples(
+                int(n_samples), domain=self.JDist, rule=sampling_method
+                )
         except:
-            samples = self.JDist.resample(n_samples)
+            samples = self.JDist.resample(int(n_samples))
 
         # Transform samples to the original space
         if transform:
@@ -84,7 +179,7 @@ class ExpDesigns:
             return samples.T
 
     # -------------------------------------------------------------------------
-    def generate_ED(self, n_samples, sample_method='random', transform=False,
+    def generate_ED(self, n_samples, sampling_method='random', transform=False,
                     max_pce_deg=None):
         """
         Generates experimental designs (training set) with the given method.
@@ -93,12 +188,12 @@ class ExpDesigns:
         ----------
         n_samples : int
             Number of requested training points.
-        sample_method : string, optional
-            Sampling method. The default is 'random'.
+        sampling_method : str, optional
+            Sampling method. The default is `'random'`.
         transform : bool, optional
-            Isoprobabilistic transformation. The default is False.
-        max_pce_deg : TYPE, optional
-            Maximum PCE polynomial degree. The default is None.
+            Isoprobabilistic transformation. The default is `False`.
+        max_pce_deg : int, optional
+            Maximum PCE polynomial degree. The default is `None`.
 
         Returns
         -------
@@ -129,25 +224,25 @@ class ExpDesigns:
             for i in range(self.ndim):
                 low_bound = np.min(Inputs.Marginals[i].input_data)
                 up_bound = np.max(Inputs.Marginals[i].input_data)
-                Inputs.Marginals[i].Parameters = [low_bound, up_bound]
+                Inputs.Marginals[i].parameters = [low_bound, up_bound]
 
         # Generate the samples based on requested method
         self.raw_data, self.bound_tuples = self.init_param_space(max_pce_deg)
 
         # Pass user-defined samples as ED
-        if self.sampling_method == 'user':
+        if sampling_method == 'user':
             samples = self.X
-            self.NrSamples = len(samples)
+            self.n_samples = len(samples)
 
         # Sample the distribution of parameters
         elif self.input_data_given:
             # Case II: Input values are directly given by the user.
 
-            if self.sampling_method == 'random':
+            if sampling_method == 'random':
                 samples = self.random_sampler(n_samples)
 
-            elif self.sampling_method == 'PCM' or \
-                    self.sampling_method == 'LSCM':
+            elif sampling_method == 'PCM' or \
+                    sampling_method == 'LSCM':
                 samples = self.pcm_sampler(max_pce_deg)
 
             else:
@@ -155,14 +250,14 @@ class ExpDesigns:
                 try:
                     samples = chaospy.generate_samples(n_samples,
                                                        domain=self.JDist,
-                                                       rule=sample_method).T
+                                                       rule=sampling_method).T
                 except:
                     samples = self.JDist.sample(n_samples)
 
         elif not self.input_data_given:
             # Case I = User passed known distributions
             samples = chaospy.generate_samples(n_samples, domain=self.JDist,
-                                               rule=sample_method).T
+                                               rule=sampling_method).T
 
         # Transform samples to the original space
         if transform:
@@ -179,7 +274,7 @@ class ExpDesigns:
         Parameters
         ----------
         max_deg : int, optional
-            Maximum degree. The default is None.
+            Maximum degree. The default is `None`.
 
         Returns
         -------
@@ -223,9 +318,7 @@ class ExpDesigns:
                                                      domain=self.JDist)
 
         # Create orthogonal polynomial coefficients if necessary
-        if self.meta_Model.lower() != 'gpe' and self.apce\
-            and max_deg is not None \
-                and Inputs.polycoeffsFlag:
+        if self.apce and max_deg is not None and Inputs.poly_coeffs_flag:
             self.polycoeffs = {}
             for parIdx in tqdm(range(ndim), ascii=True,
                                desc="Computing orth. polynomial coeffs"):
@@ -284,16 +377,11 @@ class ExpDesigns:
             if Inputs.Marginals[parIdx].dist_type is None:
                 data = Inputs.Marginals[parIdx].input_data
                 all_data.append(data)
-                dist_type, params = self.FitDist(data)
-                Inputs.Marginals[parIdx].dist_type = dist_type
-                Inputs.Marginals[parIdx].parameters = params
+                dist_type = None
             else:
                 dist_type = Inputs.Marginals[parIdx].dist_type
                 params = Inputs.Marginals[parIdx].parameters
 
-            # Make disttpe lower case for consistancy
-            dist_type = dist_type.lower()
-
             if rosenblatt:
                 polytype = 'hermite'
                 dist = chaospy.Normal()
@@ -302,36 +390,37 @@ class ExpDesigns:
                 polytype = 'arbitrary'
                 dist = None
 
-            elif 'unif' in dist_type:
+            elif 'unif' in dist_type.lower():
                 polytype = 'legendre'
                 dist = chaospy.Uniform(lower=params[0], upper=params[1])
 
-            elif 'norm' in dist_type:
+            elif 'norm' in dist_type.lower() and \
+                 'log' not in dist_type.lower():
                 polytype = 'hermite'
                 dist = chaospy.Normal(mu=params[0], sigma=params[1])
 
-            elif 'gamma' in dist_type:
+            elif 'gamma' in dist_type.lower():
                 polytype = 'laguerre'
                 dist = chaospy.Gamma(shape=params[0],
                                      scale=params[1],
                                      shift=params[2])
 
-            elif 'beta' in dist_type:
+            elif 'beta' in dist_type.lower():
                 polytype = 'jacobi'
                 dist = chaospy.Beta(alpha=params[0], beta=params[1],
                                     lower=params[2], upper=params[3])
 
-            elif 'lognorm' in dist_type:
+            elif 'lognorm' in dist_type.lower():
                 polytype = 'hermite'
                 Mu = np.log(params[0]**2/np.sqrt(params[0]**2 + params[1]**2))
                 Sigma = np.sqrt(np.log(1 + params[1]**2 / params[0]**2))
                 dist = chaospy.LogNormal(mu=Mu, sigma=Sigma)
 
-            elif 'expon' in dist_type:
+            elif 'expon' in dist_type.lower():
                 polytype = 'arbitrary'
                 dist = chaospy.Exponential(scale=params[0], shift=params[1])
 
-            elif 'weibull' in dist_type:
+            elif 'weibull' in dist_type.lower():
                 polytype = 'arbitrary'
                 dist = chaospy.Weibull(shape=params[0], scale=params[1],
                                        shift=params[2])
@@ -354,10 +443,10 @@ class ExpDesigns:
             # Naive approach: Fit a gaussian kernel to the provided data
             Data = np.asarray(all_data)
             orig_space_dist = st.gaussian_kde(Data)
-            self.priorSpace = orig_space_dist
+            self.prior_space = orig_space_dist
         else:
             orig_space_dist = chaospy.J(*orig_joints)
-            self.priorSpace = st.gaussian_kde(orig_space_dist.sample(10000))
+            self.prior_space = st.gaussian_kde(orig_space_dist.sample(10000))
 
         return orig_space_dist, poly_types
 
@@ -407,7 +496,7 @@ class ExpDesigns:
 
         raw_data = self.raw_data
 
-        # Guess the closest degree to self.NrSamples
+        # Guess the closest degree to self.n_samples
         def M_uptoMax(deg):
             result = []
             for d in range(1, deg+1):
@@ -415,7 +504,7 @@ class ExpDesigns:
                               (math.factorial(self.ndim) * math.factorial(d)))
             return np.array(result)
 
-        guess_Deg = np.where(M_uptoMax(max_deg) > self.NrSamples)[0][0]
+        guess_Deg = np.where(M_uptoMax(max_deg) > self.n_samples)[0][0]
 
         c_points = np.zeros((guess_Deg+1, self.ndim))
 
@@ -458,7 +547,7 @@ class ExpDesigns:
         if self.sampling_method.lower() == 'lscm':
             opt_col_points = sort_unique_combos
         else:
-            opt_col_points = sort_unique_combos[0:self.NrSamples]
+            opt_col_points = sort_unique_combos[0:self.n_samples]
 
         return opt_col_points
 
@@ -470,12 +559,12 @@ class ExpDesigns:
 
         Parameters
         ----------
-        X : ndarray of shape (n_samples,n_params)
+        X : array of shape (n_samples,n_params)
             Samples to be transformed.
 
         Returns
         -------
-        tr_X: ndarray of shape (n_samples,n_params)
+        tr_X: array of shape (n_samples,n_params)
             Transformed samples.
 
         """
@@ -539,7 +628,7 @@ class ExpDesigns:
         return tr_X
 
     # -------------------------------------------------------------------------
-    def FitDist(self, y):
+    def fit_dist(self, y):
         """
         Fits the known distributions to the data.
 
