@@ -84,6 +84,9 @@ class SeqDesign():
         if not isinstance(MetaModel.ExpDesign.util_func, list):
             util_func = [MetaModel.ExpDesign.util_func]
 
+        # Read observations or MCReference
+        if len(Model.observations) != 0:
+            self.observations = self.Model.read_observation()
         # ---------- Initial PCEModel ----------
         PCEModel = MetaModel.train_norm_design(Model)
         initPCEModel = deepcopy(PCEModel)
@@ -92,7 +95,7 @@ class SeqDesign():
         OutputName = Model.Output.names
 
         # Estimation of the integral via Monte Varlo integration
-        obs_data = PCEModel.Observations
+        obs_data = self.observations
 
         # Check if data is provided
         TotalSigma2 = np.empty((0, 1))
@@ -100,7 +103,7 @@ class SeqDesign():
             # ------ Prepare diagonal enteries for co-variance matrix ---------
             for keyIdx, key in enumerate(Model.Output.names):
                 # optSigma = 'B'
-                sigma2 = np.array(PCEModel.Discrepancy.Parameters[key])
+                sigma2 = np.array(PCEModel.Discrepancy.parameters[key])
                 TotalSigma2 = np.append(TotalSigma2, sigma2)
 
             # Calculate the initial BME
@@ -223,7 +226,7 @@ class SeqDesign():
                     # ------ Plot the surrogate model vs Origninal Model ------
                     if hasattr(PCEModel, 'adapt_verbose') and \
                        PCEModel.adapt_verbose:
-                        from PostProcessing.adaptPlot import adaptPlot
+                        from post_processing.adaptPlot import adaptPlot
                         y_hat, std_hat = PCEModel.eval_metamodel(samples=Xnew)
                         adaptPlot(PCEModel, Ynew, y_hat, std_hat, plotED=False)
 
@@ -250,15 +253,15 @@ class SeqDesign():
                     Yprev = PCEModel.ModelOutputDict
 
                     # Pass the new prior as the input
-                    PCEModel.input_obj.polycoeffsFlag = False
+                    PCEModel.input_obj.poly_coeffs_flag = False
                     if updatedPrior is not None:
-                        PCEModel.Inputs.polycoeffsFlag = True
+                        PCEModel.input_obj.poly_coeffs_flag = True
                         print("updatedPrior:", updatedPrior.shape)
                         # Arbitrary polynomial chaos
                         for i in range(updatedPrior.shape[1]):
-                            PCEModel.Inputs.Marginals[i].DistType = None
+                            PCEModel.Inputs.Marginals[i].dist_type = None
                             x = updatedPrior[:, i]
-                            PCEModel.Inputs.Marginals[i].InputValues = x
+                            PCEModel.Inputs.Marginals[i].raw_data = x
 
                     prevPCEModel = PCEModel
                     PCEModel = PCEModel.train_norm_design(Model)
@@ -472,8 +475,8 @@ class SeqDesign():
         Y_mean_can, Y_std_can = self.eval_metamodel(samples=np.array([X_can]))
 
         # Get the data
-        obs_data = self.Observations
-        nObs = self.Model.nObs
+        obs_data = self.observations
+        nObs = self.Model.n_obs
         # TODO: Analytical DKL
         # Sample a distribution for a normal dist
         # with Y_mean_can as the mean and Y_std_can as std.
@@ -673,7 +676,7 @@ class SeqDesign():
         PCEModel.ExpDesign.Y = NewExpDesignY
 
         # Create the SparseBayes-based PCE metamodel:
-        PCEModel.Inputs.polycoeffsFlag = False
+        PCEModel.Inputs.__poly_coeffs_flag = False
         univ_p_val = self.MetaModel.univ_basis_vals(X_can)
         G_n_m_all = np.zeros((len(Model.Output.names), Model.nObs))
 
@@ -716,7 +719,7 @@ class SeqDesign():
         PCE_SparseBayes_can = PCEModel
 
         # Get the data
-        obs_data = self.MetaModel.Observations
+        obs_data = self.observations
 
         # ---------- Inner MC simulation for computing Utility Value ----------
         # Estimation of the integral via Monte Varlo integration
@@ -934,19 +937,19 @@ class SeqDesign():
         """
 
         Model = self.Model
-        MaxFunItr = self.MetaModel.ExpDesign.MaxFunItr
+        max_func_itr = self.MetaModel.ExpDesign.max_func_itr
 
         if method == 'VarOptDesign':
             Res_Global = opt.dual_annealing(self.util_VarBasedDesign,
                                             bounds=Bounds,
                                             args=(Model, var),
-                                            maxfun=MaxFunItr)
+                                            maxfun=max_func_itr)
 
         elif method == 'BayesOptDesign':
             Res_Global = opt.dual_annealing(self.util_BayesianDesign,
                                             bounds=Bounds,
                                             args=(Model, sigma2Dict, var),
-                                            maxfun=MaxFunItr)
+                                            maxfun=max_func_itr)
 
         if verbose:
             print(f"global minimum: xmin = {Res_Global.x}, "
@@ -1176,8 +1179,8 @@ class SeqDesign():
             explore = Exploration(PCEModel, n_candidates)
             explore.w = 100  # * ndim #500
             # Select criterion (mc-intersite-proj-th, mc-intersite-proj)
-            explore.mcCriterion = 'mc-intersite-proj'
-            allCandidates, scoreExploration = explore.getExplorationSamples()
+            explore.mc_criterion = 'mc-intersite-proj'
+            allCandidates, scoreExploration = explore.get_exploration_samples()
 
             # Temp: ---- Plot all candidates -----
             if ndim == 2:
@@ -1288,7 +1291,7 @@ class SeqDesign():
                     X_can = explore.closestPoints[idx]
 
                     # Calculate the maxmin score for the region of interest
-                    newSamples, maxminScore = explore.getMCSamples(X_can)
+                    newSamples, maxminScore = explore.get_mc_samples(X_can)
 
                     # select the requested number of samples
                     Xnew[i] = newSamples[np.argmax(maxminScore)]
@@ -1310,12 +1313,6 @@ class SeqDesign():
             # Generate candidate samples from Exploration class
             nMeasurement = old_EDY[OutputNames[0]].shape[1]
 
-            # Find indices of the Vornoi cells with samples
-            goodSampleIdx = []
-            for idx in range(len(explore.closestPoints)):
-                if len(explore.closestPoints[idx]) != 0:
-                    goodSampleIdx.append(idx)
-
             # Find sensitive region
             if UtilMethod == 'LOOCV':
                 LCerror = PCEModel.LCerror
@@ -1340,7 +1337,12 @@ class SeqDesign():
                                                 axis=0)
                     goodSampleIdx = range(n_cand_groups)
                 else:
-                    split_cand = explore.closestPoints
+                    # Find indices of the Vornoi cells with samples
+                    goodSampleIdx = []
+                    for idx in range(len(explore.closest_points)):
+                        if len(explore.closest_points[idx]) != 0:
+                            goodSampleIdx.append(idx)
+                    split_cand = explore.closest_points
 
                 # Split the candidates in groups for multiprocessing
                 args = []
@@ -1357,8 +1359,12 @@ class SeqDesign():
                 pool.close()
 
                 # Retrieve the results and append them
-                ExploitScore = np.concatenate([results[NofE][1] for NofE in
-                                               range(len(goodSampleIdx))])
+                if explore_method == 'Voronoi':
+                    ExploitScore = [np.mean(results[k][1]) for k in
+                                    range(len(goodSampleIdx))]
+                else:
+                    ExploitScore = np.concatenate(
+                        [results[k][1] for k in range(len(goodSampleIdx))])
 
             else:
                 raise NameError('The requested utility function is not '
@@ -1383,12 +1389,12 @@ class SeqDesign():
                 Xnew = allCandidates[bestIdx]
             else:
                 for i, idx in enumerate(bestIdx.flatten()):
-                    X_can = explore.closestPoints[idx]
+                    X_can = explore.closest_points[idx]
                     # plotter(self.ExpDesign.X, X_can, explore_method,
                     # scoreExploration=None)
 
                     # Calculate the maxmin score for the region of interest
-                    newSamples, maxminScore = explore.getMCSamples(X_can)
+                    newSamples, maxminScore = explore.get_mc_samples(X_can)
 
                     # select the requested number of samples
                     Xnew[i] = newSamples[np.argmax(maxminScore)]
