@@ -1,16 +1,6 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-"""
-Author: Farid Mohammadi, M.Sc.
-E-Mail: farid.mohammadi@iws.uni-stuttgart.de
-Department of Hydromechanics and Modelling of Hydrosystems (LH2)
-Institute for Modelling Hydraulic and Environmental Systems (IWS), University
-of Stuttgart www.iws.uni-stuttgart.de/lh2/
-Pfaffenwaldring 61
-70569 Stuttgart
 
-Created on Sat Aug 24 2019
-"""
 import warnings
 import numpy as np
 import math
@@ -34,18 +24,71 @@ from .reg_fast_ard import RegressionFastARD
 from .reg_fast_laplace import RegressionFastLaplace
 from .bayes_linear import VBLinearRegression, EBLinearRegression
 warnings.filterwarnings("ignore")
-#plt.style.use('../bayesvalidrox.mplstyle')
+# Load the mplstyle
+plt.style.use(os.path.join(os.path.split(__file__)[0],
+                           '../', 'bayesvalidrox.mplstyle'))
 
 
 class MetaModel:
-    """ Meta (surrogate) model
+    """
+    Meta (surrogate) model
+
     This class trains a surrogate model. It accepts an input object (input_obj)
-    containing the distribution specification of uncertain parameters and a
-    model object with instructions on how to run the computational model.
-    Two surrogate model types are supported: polynomial chaos expansion (`PCE`)
-    and Gaussian process regression (`GPE`).
-    Additionally, two training modes are supported: one-shot (`normal`) and
-     adaptive sequential (`sequential`) experimental design.
+    containing the specification of the distributions for uncertain parameters
+    and a model object with instructions on how to run the computational model.
+
+    Attributes
+    ----------
+    input_obj : obj
+        Input object with the information on the model input parameters.
+    meta_model_type : str
+        Surrogate model types. Three surrogate model types are supported:
+        polynomial chaos expansion (`PCE`), arbitrary PCE (`aPCE`) and
+        Gaussian process regression (`GPE`). Default is PCE.
+    pce_reg_method : str
+        PCE regression method to compute the coefficients. The following
+        regression methods are available:
+
+        1. OLS: Ordinary Least Square method
+        2. BRR: Bayesian Ridge Regression
+        3. LARS: Least angle regression
+        4. ARD: Bayesian ARD Regression
+        5. FastARD: Fast Bayesian ARD Regression
+        6. VBL: Variational Bayesian Learning
+        7. EBL: Emperical Bayesian Learning
+        Default is `OLS`.
+    pce_deg : int or list of int
+        Polynomial degree(s). If a list is given, an adaptive algorithm is used
+        to find the best degree with the lowest Leave-One-Out cross-validation
+        (LOO) error (or the highest score=1-LOO). Default is `1`.
+    pce_q_norm : float
+        Hyperbolic (or q-norm) truncation for multi-indices of multivariate
+        polynomials. Default is `1.0`.
+    dim_red_method : str
+        Dimensionality reduction method for the output space. The available
+        method is based on principal component analysis (PCA). The Default is
+        `'no'`. There are two ways to select number of components: use
+        percentage of the explainable variance threshold (between 0 and 100)
+        (Option A) or direct prescription of components' number (Option B):
+
+            >>> MetaModelOpts.dim_red_method = 'PCA'
+            >>> MetaModelOpts.var_pca_threshold = 99.999  # Option A
+            >>> MetaModelOpts.n_pca_components = 12 # Option B
+
+    verbose : bool
+        Prints summary of the regression results. Default is `False`.
+
+    Note
+    -------
+    To define the sampling methods and the training set, an experimental design
+    instance shall be defined. This can be done by:
+
+    >>> MetaModelOpts.add_ExpDesign()
+
+    Two experimental design schemes are supported: one-shot (`normal`) and
+    adaptive sequential (`sequential`) designs.
+    For experimental design refer to `ExpDesigns`.
+
     """
 
     def __init__(self, input_obj, meta_model_type='PCE', pce_reg_method='OLS',
@@ -68,12 +111,12 @@ class MetaModel:
 
         Parameters
         ----------
-        Model : object
+        Model : obj
             Model object.
 
         Returns
         -------
-        metamodel : object
+        metamodel : obj
             The meta model object.
 
         """
@@ -85,12 +128,12 @@ class MetaModel:
             if type(self.pce_deg) is not np.ndarray:
                 self.pce_deg = np.array(self.pce_deg)
 
-        if self.ExpDesign.Method == 'sequential':
+        if self.ExpDesign.method == 'sequential':
             from .sequential_design import SeqDesign
             seq_design = SeqDesign(self)
             metamodel = seq_design.train_seq_design(Model)
 
-        elif self.ExpDesign.Method == 'normal':
+        elif self.ExpDesign.method == 'normal':
             self.ExpDesignFlag = 'normal'
             metamodel = self.train_norm_design(Model)
 
@@ -112,14 +155,14 @@ class MetaModel:
 
         Parameters
         ----------
-        Model : object
+        Model : obj
             Model object.
         verbose : bool, optional
             Flag for a sequential design in silent mode. The default is False.
 
         Returns
         -------
-        self: object
+        self: obj
             Meta-model object.
 
         """
@@ -139,11 +182,6 @@ class MetaModel:
         self.LCerror = self.auto_vivification()
         self.x_scaler = {}
 
-        # Read observations or MCReference
-        if self.ExpDesignFlag != 'sequential':
-            if len(Model.observations) != 0:
-                self.Observations = Model.read_observation()
-
         # Define the DegreeArray
         nSamples, ndim = CollocationPoints.shape
         self.DegreeArray = self.__select_degree(ndim, nSamples)
@@ -237,7 +275,7 @@ class MetaModel:
 
         Returns
         -------
-        basis_indices : array (n_terms, n_params)
+        basis_indices : array of shape (n_terms, n_params)
             Multi-indices of multivariate polynomials.
 
         """
@@ -268,7 +306,7 @@ class MetaModel:
 
         Parameters
         ----------
-        Model : object
+        Model : obj
             Model object.
 
         Raises
@@ -278,7 +316,7 @@ class MetaModel:
 
         Returns
         -------
-        ED_X_tr: array (n_samples, n_params)
+        ED_X_tr: array of shape (n_samples, n_params)
             Training samples transformed by an isoprobabilistic transformation.
         ED_Y: dict
             Model simulations (target) for all outputs.
@@ -337,7 +375,7 @@ class MetaModel:
         self.bound_tuples = ExpDesign.bound_tuples
 
         # ---- Run simulations at X ----
-        if not hasattr(ExpDesign, 'Y'):
+        if not hasattr(ExpDesign, 'Y') or ExpDesign.Y is None:
             print('\n Now the forward model needs to be run!\n')
             ED_Y, up_ED_X = Model.run_model_parallel(ED_X)
             ExpDesign.X = up_ED_X
@@ -363,13 +401,12 @@ class MetaModel:
         samples : array of shape (n_samples, n_params)
             Samples.
         n_max : int, optional
-            Maximum polynomial degree. The default is None.
+            Maximum polynomial degree. The default is `None`.
 
         Returns
         -------
-        univ_basis: array of (n_samples, n_params, n_max+1)
+        univ_basis: array of shape (n_samples, n_params, n_max+1)
             All univariate regressors up to n_max.
-
         """
         # Extract information
         poly_types = self.ExpDesign.poly_types
@@ -399,7 +436,7 @@ class MetaModel:
         basis_indices : array of shape (n_terms, n_params)
             Multi-indices of multivariate polynomials.
         univ_p_val : array of (n_samples, n_params, n_max+1)
-            All univariate regressors up to n_max.
+            All univariate regressors up to `n_max`.
 
         Raises
         ------
@@ -453,19 +490,19 @@ class MetaModel:
 
         Parameters
         ----------
-        X : array-like of shape (n_samples, n_features)
+        X : array of shape (n_samples, n_features)
             Training vector, where n_samples is the number of samples and
             n_features is the number of features.
-        y : array-like of shape (n_samples,)
+        y : array of shape (n_samples,)
             Target values.
-        basis_indices : array-like of shape (n_terms, n_params)
+        basis_indices : array of shape (n_terms, n_params)
             Multi-indices of multivariate polynomials.
-        reg_method : string, optional
+        reg_method : str, optional
             DESCRIPTION. The default is None.
 
         Returns
         -------
-        returnOuts : Dict
+        return_out_dict : Dict
             Fitted estimator, spareMulti-Index, sparseX and coefficients.
 
         """
@@ -578,12 +615,12 @@ class MetaModel:
                 coeffs = np.dot(np.dot(np.linalg.pinv(X_T_X), sparse_X.T), y)
 
         # Create a dict to pass the outputs
-        returnOuts = dict()
-        returnOuts['clf_poly'] = clf_poly
-        returnOuts['spareMulti-Index'] = sparse_basis_indices
-        returnOuts['sparePsi'] = sparse_X
-        returnOuts['coeffs'] = coeffs
-        return returnOuts
+        return_out_dict = dict()
+        return_out_dict['clf_poly'] = clf_poly
+        return_out_dict['spareMulti-Index'] = sparse_basis_indices
+        return_out_dict['sparePsi'] = sparse_X
+        return_out_dict['coeffs'] = coeffs
+        return return_out_dict
 
     # --------------------------------------------------------------------------------------------------------
     def adaptive_regression(self, ED_X, ED_Y, varIdx, verbose=False):
@@ -593,9 +630,9 @@ class MetaModel:
 
         Parameters
         ----------
-        ED_X : array-like of shape (n_samples, n_params)
+        ED_X : array of shape (n_samples, n_params)
             Experimental design.
-        ED_Y : array-like of shape (n_samples,)
+        ED_Y : array of shape (n_samples,)
             Target values, i.e. simulation results for the Experimental design.
         varIdx : int
             Index of the output.
@@ -783,31 +820,33 @@ class MetaModel:
     # -------------------------------------------------------------------------
     def corr_loocv_error(self, clf, psi, coeffs, y):
         """
-        Calculates the corrected LOO error for the OLS regression on regressor
-        matrix PSI that generated the coefficients.
-        (based on Blatman, 2009 (PhD Thesis), pg. 115-116).
+        Calculates the corrected LOO error for regression on regressor
+        matrix `psi` that generated the coefficients based on [1] and [2].
 
-        This is based on the following paper:
-           ""Blatman, G., & Sudret, B. (2011). Adaptive sparse polynomial
-           chaos expansion based on least angle regression.
-           Journal of Computational Physics, 230(6), 2345-2367.""
+        [1] Blatman, G., 2009. Adaptive sparse polynomial chaos expansions for
+            uncertainty propagation and sensitivity analysis (Doctoral
+            dissertation, Clermont-Ferrand 2).
+
+        [2] Blatman, G. and Sudret, B., 2011. Adaptive sparse polynomial chaos
+            expansion based on least angle regression. Journal of computational
+            Physics, 230(6), pp.2345-2367.
 
         Parameters
         ----------
         clf : object
             Fitted estimator.
-        psi : array-like of shape (n_samples, n_features)
+        psi : array of shape (n_samples, n_features)
             The multivariate orthogonal polynomials (regressor).
         coeffs : array-like of shape (n_features,)
             Estimated cofficients.
-        y : array-like of shape (n_samples,)
+        y : array of shape (n_samples,)
             Target values.
 
         Returns
         -------
         Q_2 : float
             LOOCV Validation score (1-LOOCV erro).
-        residual : array-like of shape (n_samples,)
+        residual : array of shape (n_samples,)
             Residual values (y - predicted targets).
 
         """
@@ -893,14 +932,14 @@ class MetaModel:
 
         Parameters
         ----------
-        Output : array-like of shape (n_samples,)
+        Output : array of shape (n_samples,)
             Target values.
 
         Returns
         -------
-        pca : object
+        pca : obj
             Fitted sklearnPCA object.
-        OutputMatrix : array-like of shape (n_samples,)
+        OutputMatrix : array of shape (n_samples,)
             Transformed target values.
 
         """
@@ -944,7 +983,7 @@ class MetaModel:
         return pca, OutputMatrix
 
     # -------------------------------------------------------------------------
-    def gaussian_process_emulator(self, X, y, nugTerm=None, autoSelect=False,
+    def gaussian_process_emulator(self, X, y, nug_term=None, autoSelect=False,
                                   varIdx=None):
         """
         Fits a Gaussian Process Emulator to the target given the training
@@ -952,11 +991,11 @@ class MetaModel:
 
         Parameters
         ----------
-        X : array-like of shape (n_samples, n_params)
+        X : array of shape (n_samples, n_params)
             Training points.
-        y : array-like of shape (n_samples,)
+        y : array of shape (n_samples,)
             Target values.
-        nugTerm : float, optional
+        nug_term : float, optional
             Nugget term. The default is None, i.e. variance of y.
         autoSelect : bool, optional
             Loop over some kernels and select the best. The default is False.
@@ -970,15 +1009,15 @@ class MetaModel:
 
         """
 
-        nugTerm = nugTerm if nugTerm else np.var(y)
+        nug_term = nug_term if nug_term else np.var(y)
 
-        Kernels = [nugTerm * kernels.RBF(length_scale=1.0,
-                                         length_scale_bounds=(1e-25, 1e15)),
-                   nugTerm * kernels.RationalQuadratic(length_scale=0.2,
-                                                       alpha=1.0),
-                   nugTerm * kernels.Matern(length_scale=1.0,
-                                            length_scale_bounds=(1e-15, 1e5),
-                                            nu=1.5)]
+        Kernels = [nug_term * kernels.RBF(length_scale=1.0,
+                                          length_scale_bounds=(1e-25, 1e15)),
+                   nug_term * kernels.RationalQuadratic(length_scale=0.2,
+                                                        alpha=1.0),
+                   nug_term * kernels.Matern(length_scale=1.0,
+                                             length_scale_bounds=(1e-15, 1e5),
+                                             nu=1.5)]
 
         # Automatic selection of the kernel
         if autoSelect:
@@ -1028,7 +1067,7 @@ class MetaModel:
         nsamples : int, optional
             Number of samples to generate, if no `samples` is provided. The
             default is None.
-        sampling_method : string, optional
+        sampling_method : str, optional
             Type of sampling, if no `samples` is provided. The default is
             'random'.
         return_samples : bool, optional
@@ -1036,9 +1075,10 @@ class MetaModel:
 
         Returns
         -------
-        TYPE
-            DESCRIPTION.
-
+        mean_pred : dict
+            Mean of the predictions.
+        std_pred : dict
+            Standard deviatioon of the predictions.
         """
         if self.meta_model_type.lower() == 'gpe':
             model_dict = self.gp_poly
@@ -1122,13 +1162,13 @@ class MetaModel:
 
         Parameters
         ----------
-        X : array-like of shape (n_outputs, n_inputs)
+        X : array of shape (n_outputs, n_inputs)
             Input array. It can contain any forcing inputs or coordinates of
              extracted data.
-        y : array-like of shape (n_outputs,)
+        y : array of shape (n_outputs,)
             The model response for the MAP parameter set.
-        name : string, optional
-            Calibration or validation. The default is 'Calib'.
+        name : str, optional
+            Calibration or validation. The default is `'Calib'`.
 
         Returns
         -------
@@ -1215,7 +1255,11 @@ class MetaModel:
 
     # -------------------------------------------------------------------------
     class auto_vivification(dict):
-        """Implementation of perl's AutoVivification feature."""
+        """
+        Implementation of perl's AutoVivification feature.
+
+        Source: https://stackoverflow.com/a/651879/18082457
+        """
 
         def __getitem__(self, item):
             try:
